/* ---------------------------------------------------------------------------------------------
 *										R.e.e.f.
 * ---------------------------------------------------------------------------------------------
 *
 *	projeto: dHome
 *	autores: Adelmario Douglas Leite Cabral Junior - adlcj[at]cin.ufpe.br
 *			 Lamberto Augusto De Oliveira Neto - laon[at]cin.ufpe.br
 *			 Marcelo Bezerra de Amorim - mba3[at]cin.ufpe.br
 *			 Francisco de Farias Dantas - ffd[at]cin.ufpe.br	
 *
 *  mdulo: PLC
 *  ultima atualizao: 26/11/2005
*/


#include <REG51.H>

/*
 * Constantes para configurao da comunicao na serial
 */				  

//Valores definidos para baud rate da Serial
#define CT_CLK_FREQ 12000000
#define CT_BAUD 4800
#define CT_TIMER1_RELOAD ( 256 - ( ( CT_CLK_FREQ / (384) ) / CT_BAUD ) )

#define BAUD0 1200
#define TIMER_RELOAD0 ( 256 - ( ( CT_CLK_FREQ / (384) ) / BAUD0 ) )

//Estados principais do Mdulo PLC
#define CT_WAITING0 0					//Estado 'Desativado0'
#define CT_WAITING1 1					//Estado 'Desativado1'
#define CT_RUNNING 2					//Estado 'Funcionando'

//Defines gerais
#define CT_START0 0x0F
#define CT_START1 0xAA


//Defines referente ao controlador
#define CT_ACK 0x06 						// corresponde ao estado que envia todo o comando X10
#define CT_OK 0xFF							// corresponde ao estado que indica que o comando foi enviado com sucesso
#define CT_FETCH 0xFE						// corresponde ao estado que recebe os dados do servidor
#define CT_START 0x02						// corresponde ao estado que envia o cdigo de iniciar comando para o dispositivo X10
#define CT_NAK 0x15							// corresponde ao estado que indica que o cdigo no foi identificado
#define CT_CR 0x0D							// corresponde ao cdigo de retorno
#define CT_ON 0xFD							// corresponde ao estado que verifica se  para enviar comando ou desativar conexo 
#define CT_NOP 0x99							// corresponde ao estado para iniciar conexo com o controlador X10
#define CT_ERRO 0x97						// corresponde ao estado que indica algum erro ocorrido 
#define CT_send_x10_command 0x63			// corresponde ao cdigo X10 de enviar comando


//defines da serial emulada
#define START_SERIAL 0
#define STOP_SERIAL 1

//Bits de controle do handshake (a definir)
sbit STROBLE = P2^0;					//Sinal Stroble da paralela
sbit ACK_P = P2^1;						//Sinal ACK para a paralela

//Variveis
char data0;
char data1;
char data2;


//Variveis do Controlador

static char KEY;							// chave usada na mquina de estados
char extended[64];							// varivel que armazena os dados extendido recebidos do servidor
char house_code_s;							// varivel que armazena o cdigo da casa recebido do servidor
char unit_code_s;							// varivel que armazena o cdigo do dispositivo recebido do servidor
char function_code_s;						// varivel que armazena o cdigo da funo recebido do servidor
char extra_code_s;							// varivel que armazena a quantidade da dimerizao
char dados = '0';							// varivel que auxilia no envio do comando 
char tmp, tmp2, tmp3, tmp4, tmp5;			// variveis que auxiliam no envio do comando
char tmp6, tmp7, tmp8,tmp9, tmp10, tmp11;	// variveis que auxiliam no envio do comando
 			  

int i = 0; 									//varivel que armazena a quantidade de bytes do cdigo extendido

enum {ON, OFF, FETCH, START, ACK, ERRO, OK, NAK};


//variveis da serial emulada
enum {WAITING, TRANSMITING, FINALIZING};                     //Estados de Transmio de Dados via serial emulada
enum {WAITING_RECEIVE, RECEIVING, FINALIZING_RECEIVE};       //Estados da Recepo de Dados via serial emulada

sbit sTx = P1^0;
sbit sRx = P1^1;

bit GTI = 0;
bit GRI = 0;

bit send = 0;
bit receive = 0;

char time = 2;
char timeR= 1;	//Botei timeR pq num sabia se pode trasmitir e receber ao mesmo tempo

char n = 7;
char nR = 00000001;
char SNBUF;
char RCBUF;


void timer0ISR() interrupt 1 using 0{
	static char sendState = WAITING;
	static char receiveState = WAITING_RECEIVE;

	TR0 = 0;
	TL0 = TIMER_RELOAD0;
	TR0 = 1;

	if(send){

		switch(sendState){

			case WAITING:

				sTx = START_SERIAL;

				if(time == 0){

					time = 2;
					sendState = TRANSMITING;
					break;
				}
				time--;
				break;

			case TRANSMITING:

				if(SNBUF & 0x01){
					sTx = 1;
				} else {
					sTx = 0;
				}

				if((n == 0) && (time == 0)){

					time = 2;
					sendState = FINALIZING;
					break;

				} else if(time == 0){

					SNBUF = SNBUF >> 1;
					time = 2;
					n--;
					break;

				}
				time--;
				break;

			case FINALIZING:

				sTx = STOP_SERIAL;

				if(time == 0){

					GTI = 1;
					time = 2;
					sendState = WAITING;
					break;

				}
				time--;
				break;

		}
	}


	if(receive){

      switch(receiveState){

		case WAITING_RECEIVE :
		    //Se o valor de sRx for igual a START_SERIAL o byte deve ser recebido
		    if(sRx == START_SERIAL){
		    //Garante que a verificao ser feita duas vezes para garantir
				if (timeR == 0){
		    		receiveState = RECEIVING;
			  		timeR = 1;
			  		RCBUF = 0;
			  		GRI = 0;
				}
				else {
			  		timeR--;
				}
            } else { //Caso o START_SERIAL for recebido apenas uma vez, pode ser simplesmente ruido
		    	timeR = 1;
		    }
		   break;

	    case RECEIVING :
            //Verifica se  a primeira vez que o bit est sendo analisado
		    if(timeR){
		    	//Se o bit recebido for 1 o bit correspondente no buffer ser setado
                if(sRx){
		 	    	RCBUF = RCBUF | nR;
				}
			//Se o bit recebido for 0, o valor que j era 0 no se altera
				timeR--;
		   }

		   //Na segunda verificao, no importa o que tem no bit de recepo
		  else {
          	timeR = 1;
		    nR = nR << 1;
		    //Caso nR seja igual a zero, todos os bits j foram verificados
		    if(!nR){
               	receiveState = FINALIZING_RECEIVE;
            }
          }

         break;

		case FINALIZING_RECEIVE :
        	if(sRx == STOP_SERIAL){
				if(timeR){
		   	    	timeR--;
				} else {
			    	GRI = 1;
			    	timeR = 1;
			    	receiveState = WAITING_RECEIVE;
			    	nR = 00000001;
				}
		    } else {   // verificar se est correto depois
			    GRI = 0;
			    timeR = 1;
			    receiveState = WAITING_RECEIVE;
			    nR = 00000001;
            }
		   break;
       }
	}
}

//Timer0 mode 2 auto-reload
void initializeTimer0() {
	TR0 = 0;
	TMOD = (TMOD & 0xF0) | 0x02;
	TH0 = TIMER_RELOAD0;
	TL0 = TIMER_RELOAD0;
	ET0 = 1;
	TR0 = 1;
}


/*
*Inicializa Serial modo 1 full-duplex
*/
void initializeSerial(){
	SCON = 0x50;
}

/*
*Inicializa Timer1 modo 2 Auto-Reload
*/
void initializeTimer1(){
	TR1 = 0;							//Desliga Timer1
	TMOD = (TMOD & 0x0F) | 0x20;		//Configura para modo 2
	TH1 = CT_TIMER1_RELOAD;				//Valor de Auto-Reload
	TL1 = CT_TIMER1_RELOAD;				//Valor Inicial
	TR1 = 1;							//Liga Timer1
}

/*
 * Envia um dado pela serial
 *
 * msg - dado a ser enviado pela serial
 */

void sendToServer(char msg){
	SBUF = msg;								// envia o dado
	while(!TI);								// espera terminar de transmitir o dado
	TI = 0;									// resetando o TI
}


/*
 * Retorna um comando recebido do Servidor via serial
 */

char receiveFromServer(){
	char temp = 0;
	while(!RI);								// checa se o dado foi completamente recebido
	temp = SBUF;							// copia o dado recebido para temp
	RI = 0;
	return temp;
}



/*
 * Retorna um dado recebido dos dispositivos via Serial simulada 
 */

char receiveSerial(){
	receive = 1;
	while(!GRI);
	receive = 0;
	GRI = 0;
	return RCBUF;
}


/*
 * Envia um dado pela Serial simulada
 *
 * msn - mensagem a ser enviada
 */
void sendSerial(char msn){
	SNBUF = msn;
	send = 1;
	while(!GTI);
	send = 1;
	GTI = 0;
}



/*
*quina de estados que recebe comandos do Power Linc
*/
void receiveX10(){
	;									//Que cdigos ASCII o Power Linc manda pra que eu saiba que chegou um comando aleatorio???
}

/*
 * Intercala um bit negado de cada bit dos 4 bits menos significativos do byte passado como parmetro,
 * retornando um byte com os bits intercalados.
 *
 * src - byte com os bits menos significativos contendo o que se deseja intercalar com bits negados.
 */

char transformByte(char src){
	char ret, nsrc, and;
	int ind;
	
	nsrc = !src;							//byte que conter os bits negados a serem intercalados
	ret = 0x00;								//byte que ser retornado
	and = 0x80;							
	
    //lao que intercala os bits negados
	for(ind = 0; ind < 4; ind++){//incio do for
		ret = ret | ((src>>ind)&and);		//coloca um bit dos quatro bits originais recebidos
		and = and>>1;
		ret = ret | ((nsrc>>(ind+1))&and);	//coloca o respectivo bit negado
		and = and>>1;						
	}//fim do for

	return ret;
}


/*
 * Funo que envia o comando X10 requisitado pelo servidor
 *
 * house_code    - Cdigo da casa recebido do servidor
 * unit_code     - Cdigo do dispositivo a ser manipulado recebido do servidor 
 * function_code - Cdigo da funo a ser executada recebido do servidor
 */

void sendX10Command(char house_code, char unit_code, char function_code){
	char house_code_t, function_code_t, cnt;
	int ind;

	//verifica se a funo recebida  de dimerizao ou contm dados extendidos
	if( ((function_code_s==0x09) | (function_code_s==0x0B)) ^ (!i) ){//incio do if
		sendSerial(0x63);								//envia o comando padro de X10 
		sendSerial(house_code);							//envia o cdigo da casa
		sendSerial(unit_code);							//envia o cdigo do dispositivo
		sendSerial(function_code);						//envia a funo
		sendSerial(0x41);				//envia o repeat command (padro de X10)	
		
		tmp2 = receiveSerial();	  	
		tmp3 = receiveSerial();
		tmp4 = receiveSerial();
		tmp5 = receiveSerial();
		tmp6 = receiveSerial();
		tmp7 = receiveSerial();
		tmp8 = receiveSerial();
		tmp9 = receiveSerial();
		tmp10 = receiveSerial();
		tmp11 = receiveSerial();
	
		//house_code_s = house_code;
		//unit_code_s = unit_code;
		//function_code_s = function_code;

	}//fim do if 
	else {//incio do else	     
		cnt = (house_code_s<<4);		
		house_code_t = transformByte(cnt);								//coloca os bits intercalados do house_code_s
		function_code_t = transformByte((function_code_s<<3) & 0xFE);   //coloca os bits intercalados do function_code_s
		
		sendSerial(0x80);												// envia o cdigo padro de dados extendidos
		sendSerial(extra_code_s);										// envia a qauntidade de bytes dos dados extendidos
		
		dados = 0xE0;													// CT_START code no nibble mais significativo
		dados = (dados | (house_code_t>>4));
		sendSerial(dados);												// CT_START_code + 4 bits mais significativos de house_code com bits barrados intercalados 
		
		dados = ((0x0F & house_code_t)<<4);								// 4 bits menos significativos de house_code com bits barrados intercalados 
		dados = (dados | (function_code_t>>4));
		sendSerial(dados); 												// 4 bits menos significativos de house_code com bits barrados intercalados + 4 bits mais significativos de function_code com bits barrados intercalados 
		
		dados = (((0x0F & function_code_t)<<4) & 0xF8); 				// 4 bits menos significativos de function_code com bits barrados intercalados + 10
		dados = (dados | (extended[0]>>6)); 
		sendSerial(dados); 												// 4 bits menos significativos de function_code com bits barrados intercalados + 4 bits menos significativos do primeiro byte do dado extendindo

		//lao para enviar os dados extendidos desde do 3 bit at restarem 6 bits
		for (ind = 0; ind < ((2*i)-1); ind++){//incio do foror
			dados = ((extended[ind]<<2) & 0xFC);
			dados = (dados | ((extended[ind+1])>>6));
			sendSerial(dados);
		}//fim do for
		
		dados = ((0xFC & extended[ind])<<2);
		sendSerial(dados);												//envia os 6 bits restantes junto com 2 zeros dos 6 definidos no padro X10
		
		dados = 0x00;
		sendSerial(dados);												//envia os 4 zeros restantes do padro X10
	}//fim do else
}


/*
 * Verifica se os dados recebidos so vlidos
 */

char check(){
	char cnt;
	//Verifica se no contm dados extendidos
	if(!i){//incio do if		
		if(tmp2 != 0x58) cnt = CT_ERRO; 										//transmission received		
		if((cnt != CT_ERRO) && (tmp3 != house_code_s)) cnt = CT_ERRO;								   		
		if((cnt != CT_ERRO) && (tmp4 != unit_code_s)) cnt = CT_ERRO;								   		
		if((cnt != CT_ERRO) && (tmp5 != 0x31)) cnt = CT_ERRO; 					//sent o nmero de vezes				   		
		if((cnt != CT_ERRO) && (tmp6 != CT_CR)) cnt = CT_ERRO;										   		
																									   
		if((cnt != CT_ERRO) && (tmp7 != 0x58)) cnt = CT_ERRO; 					//transmission received		
		if((cnt != CT_ERRO) && (tmp8 != house_code_s)) cnt = CT_ERRO;							   		
		if((cnt != CT_ERRO) && (tmp9 != function_code_s)) cnt = CT_ERRO;						   		
		if((cnt != CT_ERRO) && (tmp10 != 0x31)) cnt = CT_ERRO; 					//sent once						   		
		if((cnt != CT_ERRO) && (tmp11 != CT_CR)) cnt = CT_ERRO;
	}//fim do if
	return cnt;
}


/*
 * Funo principal que contm a mquina de estados
 */

void sendX10(char b, char dt, int e){	
	char cod1, cod2;	
	int index;
	char bool = b;
	char data_tmp = dt;
	int erro = e;

			
	//mquina de estados do controlador
	switch(KEY) {//incio do switch
		case OFF://caso que verifica se h algo novo vindo do servidor
			erro = 1;

			dados = receiveFromServer();

			//verifica se  para ativar X10
			if (dados == 0xC1) {//incio do if 
				sendToServer(0x01); 			//pe ok no barramento
		 		KEY = ON; 						//vai pra CT_ON
			}//fim do if 
			else sendToServer(0x0F); 			//erro

			break;

		case ON://caso que envia comando ou destiva a conexo
			
			dados = receiveFromServer();

			// verifica se  para desativar conexo
			if(dados == 0xCF){//incio do if 
				sendToServer(0x01); 			// pe CT_OK no barramento
		 		KEY = OFF; 						// vai pra CT_NOP
			}//fim do if
			//verifica se  para enviar comando
		 	else if(dados == 0xC3){//incio do else if
				sendToServer(0x03); 			//esperando o codigo da casa
				KEY = FETCH; 					//vai para CT_FETCH
			}//fim do else if					

			break;

		case FETCH://caso que organiza os cdigos a serem enviados que esto sendo recebidos do Servidor
			house_code_s = receiveFromServer();
			sendToServer(0x04); 				//envia -> esperando cd, func. casa

			dados = receiveFromServer();

			unit_code_s = ((dados >>3 ) & 0x1E);
			function_code_s = ((dados & 0x0F) << 1) | 0x01;

			sendToServer(0x05); 				//envia -> esperando extra cd

			dados = receiveFromServer();

			extra_code_s = dados;
			i = dados - '0';					//transforma dados no tipo int

			//verifica se contm dados extendidos
			if( i ){//incio do if
				sendToServer(0x06); 			//envia -> esperando extra dados
				
				//lao para colocar no buffer os dados extendidos recebidos
				for(index=0;index<i;index++){//incio do for
					dados = receiveFromServer();						
					*(&extended+(index*2)) = transformByte(dados);
					*(&extended+((index*2)+1)) = transformByte(dados<<4);						
					//verifica se no  o ultimo byte a receber
					if(index!=i-1){//incio do if
						sendToServer(0x06); 	//envia -> esperando extra dados
					}//fim do if
				}//fim do for
			}//fim do if

			sendToServer(0x07); 				//envia -> transmisso completa

			KEY = START;						//vai para CT_START
			break;

		case START://caso que recebe o cdigo do X10 
			sendSerial(CT_START);
			cod1 = receiveSerial();					
							
			if(cod1 == CT_ACK){
			   cod2 = receiveSerial();
			   KEY = ACK;
			}
			else if(cod1 == CT_NAK) KEY = NAK;
			else KEY = ERRO;				
			break;

		case NAK://caso que envia para o servidor se houve erro
			//verifica se houve erro
			if(erro){//incio do if
				sendToServer(0x0F); 			//envia -> CT_ERRO
				KEY = START;
				erro = 0;
			}//fim do if
			else{//incio do else
				KEY = ON;
			}//fim do else
			break;

		case ACK://caso que envia o comando X10 recebido
			receiveSerial();						 					//recebe o CT_CR
			
			sendX10Command(house_code_s ,unit_code_s ,function_code_s); // envia o comando X10
			
			//verifica se deu erro
			if(check() == CT_ERRO) KEY = ERRO;													// checa se existe algum erro
			else KEY = OK;
			break;

		case ERRO://caso que trata o erro
			//verifica se houve erro
			if(erro){//incio do if
                KEY = START;
				erro = 0;
			}//fim do if
			else{//incio do else
				sendToServer(0x0F); 									//envia -> CT_ERRO
				KEY = ON;
			}//fim do else
			break;

		case OK://caso que indica que a execuo ocorreu com sucesso
			KEY = ON;
			break;
	}//fim do switch 
}

/*
*Mquina de estados principal
*Verifica e gerencia os estados do Mdulo PLC
*/
void main(){
	char bool = 0;
	char data_tmp = 0;
	int erro = 1;
	char state = CT_RUNNING;			//Estado inicial

	KEY = OFF;

	initializeTimer0();
	initializeTimer1();					//Inicializa Timer1
	initializeSerial();					//Inicializa Serial
	EA = 1;								//Ativa interrupes

	//Loop principal
	while(1){

		//Percorre os estados para inicializao e operao do Mdulo PLC
		switch(state){

			//Espera conexo do Servidor
			case CT_WAITING0:

				//Recebe Stroble = 1
				if(STROBLE){

					//Recebe 0x0F
					if(P0 == CT_START0){
						P2 = (P2 & 0x0F) | 0xA0;	//Envia o nibble 0xA
						ACK_P = 1;					//Envia ACK
						state = CT_WAITING1;
					}

				}
			
				break;

			//Confirma conexo com o Servidor
			case CT_WAITING1:			

				//Recebe Stroble = 0
				if(!STROBLE){
					
					//Recebe 0xAA
					if(P0 == CT_START1){	//Caso receba 0xAA, est ligado
						state = CT_RUNNING;
					} else {				//Caso contrrio, volta a esperar conexo do Servidor
						state = CT_WAITING0;
					}

				}

				break;

			case CT_RUNNING:					//Estado de operao do Mdulo PLC

				sendX10(bool, data_tmp, erro);	//Envia comandos para o Power Linc
				receiveX10();					//Recebe comandos do Power Linc

				break;

		}

	}

}
